/**
 * CMI : Cluster Method Invocation
 * Copyright (C) 2008 Bull S.A.S.
 * Contact: carol@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 * --------------------------------------------------------------------------
 * $Id: LoadBalancingClassLoader.java 1768 2008-05-05 10:22:54Z eyindanga $
 * --------------------------------------------------------------------------
 */

/**
 *Defines Class loader for load balancing archives
 */
package org.ow2.cmi.lb.loader;

import java.io.File;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.Iterator;
import java.util.List;
import java.util.concurrent.ConcurrentHashMap;

import org.ow2.cmi.lb.policy.IPolicy;
import org.ow2.cmi.lb.strategy.IStrategy;
import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;


/**
 * @author The new CMI team
 *
 */
public class LoadBalancingClassLoader extends URLClassLoader {

    /**
     * Logger.
     */
    private static final Log LOGGER = LogFactory.getLog(LoadBalancingClassLoader.class);

    /**
     * LoadBalancingClassLoader name
     */
    private String name = null;

    /**
     * Need to recompute toString() value ? (urls have changed)
     * True by default (not done).
     * Then, compute is done only when required and if needed
     */
    private boolean recomputeToString = true;

    /**
     * Need to recompute getClassPath() value ? (urls have changed)
     * True by default (not done).
     * Then, compute is done only when required and if needed
     */
    private boolean recomputeClassPath = true;

    /**
     * String representation used by toString() method
     */
    private String toStringValue = null;

    /**
     * Classpath value
     */
    private String classpath = null;
    /**
     * Store each class of Strategy.
     */
    private final ConcurrentHashMap<String, Class<? extends IStrategy>> strategyMap =
        new ConcurrentHashMap<String, Class<?extends IStrategy>>();

    /**
     * Store each class of Policy.
     */
    @SuppressWarnings("unchecked")
    private final ConcurrentHashMap<String, Class<? extends IPolicy>> policyMap  =
        new ConcurrentHashMap<String, Class<? extends IPolicy>>();

    /**
     * Constructs a new ClassLoader with the specified URLs.
     * @param name ClassLoader name (used for Display)
     * @param urls the URLs to add at the ClassLoader creation.
     * @param parent parent ClassLoader, null if no parent.
     */
    public LoadBalancingClassLoader(final String name, final URL[] urls, final ClassLoader parent) {
        super(urls, parent);
        this.name = name;
        this.recomputeToString = true;
        this.recomputeClassPath = true;
    }

    /**
     * Constructs a new ClassLoader with the specified URLs.
     * Uses the default delegation parent classloader.
     * @param name ClassLoader name (used for Display)
     * @param urls the URLs to add at the ClassLoader creation.
     */
    public LoadBalancingClassLoader(final String name, final URL[] urls) {
        super(urls, Thread.currentThread().getContextClassLoader());
        this.name = name;
        this.recomputeToString = true;
        this.recomputeClassPath = true;
    }

    /**
     * Add the specified URL to this ClassLoader.
     * @param url the URL to add to this ClassLoader.
     */
    @Override
    public void addURL(final URL url) {
        if (url != null) {
            super.addURL(url);
        }
        this.recomputeToString = true;
        this.recomputeClassPath = true;
    }

    /**
     * Add the specified URLs to this ClassLoader.
     * @param urls the URLs to add to this ClassLoader.
     */
    public void addURLs(final URL[] urls) {
        if (urls != null) {
            for (int i = 0; i < urls.length; i++) {
                if (urls[i] != null) {
                    super.addURL(urls[i]);
                }
            }
        }
        this.recomputeToString = true;
        this.recomputeClassPath = true;
    }

    /**
     * Display all the URLs from the class loader
     */
    public void printURLs() {
        System.out.println(name  + " ClassLoader :");
        URL[] urls = super.getURLs();
        for (int i = 0; i < urls.length; i++) {
            System.out.println("url=" + (new File(urls[i].getFile())).getAbsolutePath());
        }
        // display parent classloader
        if (getParent() != null && getParent() instanceof LoadBalancingClassLoader) {
            System.out.println("parent :");
            ((LoadBalancingClassLoader) getParent()).printURLs();
        }
    }

    /**
     * Get the class path of this classloader
     * @return the class path of this classloader
     */
    public String getClassPath() {
        // urls have changed, need to build value
        if (recomputeClassPath) {
            computeClassPath();
        }
        return classpath;
    }

    /**
     * Displays useful information
     * @return information
     */
    @Override
    public String toString() {
        // urls have changed, need to build value
        if (recomputeToString) {
            computeToString();
        }
        return toStringValue;
    }

    /**
     * Compute a string representation used by toString() method
     */
    private void computeToString() {
        StringBuffer sb = new StringBuffer();
        sb.append(this.getClass().getName());
        sb.append("[");
        sb.append(name);
        sb.append(", urls=");
        URL[] urls = getURLs();
        for (int u = 0; u < urls.length; u++) {
            sb.append(urls[u]);
            if (u != urls.length - 1) {
                sb.append(";");
            }
        }
        sb.append("]");
        toStringValue = sb.toString();

        // value is updated, no need to do it again.
        recomputeToString = false;
    }

    /**
     * Compute classpath value
     */
    private void computeClassPath() {
        String cp = "";
        // add parent classpath before
        if (getParent() != null && getParent() instanceof LoadBalancingClassLoader) {
            cp += ((LoadBalancingClassLoader) getParent()).getClassPath();
        }
        URL[] urls = super.getURLs();
        for (int i = 0; i < urls.length; i++) {
            cp = cp + File.pathSeparator + (new File(urls[i].getFile())).getAbsolutePath();
        }
        classpath = cp;
        // update value
        recomputeClassPath = false;
    }
    /**
     * Gets Load balancing prolicies and strategies from the URLs.
     */
    public void compute(final Iterator<URL> archiveIterator) {
        LOGGER.debug(" Trying to get load balacing strategies and policies from URLs '" + this.getURLs() + "'");
        try {
            while (archiveIterator.hasNext()) {
                URL url = archiveIterator.next();
                if (url.toExternalForm().endsWith(".class")) {
                    try {
                        String className = getClassName(url.toExternalForm());
                        Class<?> myclass = loadClass(className);
                        Class[] args = {List.class};
//                        try {
//                            myclass.cast(IPolicy.class);
//                            policyMap.put(className, (Class<? extends IPolicy>) myclass);
//                        } catch (Exception e) {
//                            try {
//                                myclass.cast(IStrategy.class);
//                                strategyMap.put(className, (Class<? extends IStrategy>) myclass);
//                            } catch (Exception e1) {
//
//                            }
//                        }
                        if (IPolicy.class.isInstance(myclass.getClassLoader())) {
                            LOGGER.debug("loading policy class named '" + className + "'");
                            policyMap.put(className, (Class<? extends IPolicy>) myclass);
                        }else if (IStrategy.class.isInstance(myclass)) {
                            strategyMap.put(className, (Class<? extends IStrategy>) myclass);
                        }
//                        if (isA(myclass , IPolicy.class)) {
//                            LOGGER.debug("loading policy class named '" + className + "'");
//                            policyMap.put(className, (Class<? extends IPolicy>) myclass);
//                        }else if (isA(myclass , IStrategy.class)) {
//                            strategyMap.put(className, (Class<? extends IStrategy>) myclass);
//                        }
//                        if (myclass.getMethod("getStrategy", null) != null) {
//                            LOGGER.debug("loading policy class named '" + className + "'");
//                            policyMap.put(className, (Class<? extends IPolicy>) myclass);
//
//                        }else if(myclass.getMethod("choose", args) != null) {
//                            strategyMap.put(className, (Class<? extends IStrategy>) myclass);
//                        }
                    }catch (Exception e) {
                        // TODO: handle exception
                    }
                }
            }
        } catch (Exception e) {
            // TODO: handle exception
        }

    }

    /**
     * @return the strategyMap
     */
    public ConcurrentHashMap<String, Class<? extends IStrategy>> getStrategyMap() {
        return strategyMap;
    }

    /**
     * @return the policyMap
     */
    public ConcurrentHashMap<String, Class<? extends IPolicy>> getPolicyMap() {
        return policyMap;
    }

    /**
     * Gets valid package name from url path.
     * <code>org/ow2/my_package/My_class</code> returns <code>org.ow2.my_package.My_class</code>
     * @param filePath
     * @return
     */
    private static String getClassName(final String filePath) {
        int id = filePath.lastIndexOf("!");
        String unformattedClsName = filePath.substring(id + 2 ).replace(".class", "");
        String ret = unformattedClsName.replaceAll("/", ".");
        return ret;
    }
    /**
     * Checks if the parameter class <code>someClass</code> is a subclass of class <code>superItf</code>
     * returns true if so, and else otherwise.
     * @param someClass
     * @param superItf check if
     */
    private static boolean isA(final Class<?> someClass, final Class superItf) {
        Class[] itfs = someClass.getInterfaces();
        for (int i = 0; i < itfs.length; i++) {
            if (itfs[i].getName().equals(superItf.getName())) {
                return true;
            }
        }
        return false;
    }
}
